(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)


Require Import Bool.
Require Import List.
Require Import Coq.Program.Utils.
Require Import Case.



Lemma fst_pair : forall X Y (x : X) (y : Y), fst (x, y) = x.
Proof. intros X Y x y. reflexivity. Qed.

Lemma split_cons : forall (X Y : Type) (x : X) (y : Y) (xys : list (X * Y)),
  split ((x,y) :: xys) = (x :: fst (split xys), y :: snd (split xys)).
Proof. intros. simpl. remember (split xys) as sxys. destruct sxys. reflexivity. Qed.

Lemma fst_split_cons : forall (X Y : Type) (x : X) (y : Y) (xys : list (X*Y)),
  fst (split ((x, y) :: xys)) = x :: (fst (split xys)).
Proof. intros. rewrite split_cons. reflexivity. Qed.

Lemma snd_split_cons : forall (X Y : Type) (x : X) (y : Y) (xys : list (X*Y)),
  snd (split ((x, y) :: xys)) = y :: (snd (split xys)).
Proof. intros. rewrite split_cons. reflexivity. Qed.

Lemma fst_split_nil : forall X Y (xs : list (X * Y)),
  fst (split xs) = nil -> xs = nil.
Admitted.

Lemma fst_split_eq_cons : forall X Y (xys : list (X * Y)) x xs,
  fst (split xys) = x :: xs ->
    exists y, exists xys',
      xys = (x,y) :: xys' /\
      fst (split xys') = xs.
Admitted.

Lemma in_fst_split : forall X Y (l : X) (lts : list (X * Y)),
  In l (fst (split lts)) ->
  exists T, In (l,T) lts.
Proof.
  intros. induction lts as [| [l' T'] lts']. inversion H.
  rewrite fst_split_cons in H. simpl in H. destruct H. simpl.
    subst. eexists. left. reflexivity.
    apply IHlts' in H. destruct H as [T IH].
    eexists. right. apply IH.
Qed. 

Lemma and_eq_true : forall b1 b2, (b1 && b2) = true -> b1 = true /\ b2 = true.
Proof. intros; split; destruct b1; destruct b2; auto. Qed.

Lemma not_or : forall P Q, ~P /\ ~Q -> ~(P \/ Q).
Proof. intros P Q H Hc. destruct H as [H1 H2]. destruct Hc as [H0 | H0]. apply H1. apply H0.
  apply H2. apply H0. Qed.

Lemma not_or_rev : forall P Q, ~(P \/ Q) -> ~P /\ ~Q.
Proof. intros P Q H. split; intro Hc; apply H. left. assumption. right. assumption. Qed.

Lemma eqX_dec_false : forall X (eqX_dec : forall (x y : X), {x = y}+{x <> y}) x y,
  (if eqX_dec x y then true else false) = false -> x <> y.
Proof. intros. remember (eqX_dec x y) as b. destruct b. inversion H. assumption. Qed.

Lemma eqX_dec_false_rev : forall X (eqX_dec : forall (x y : X), {x = y}+{x <> y}) x y,
  x <> y -> exists pf, eqX_dec x y = right (x = y) pf.
Proof. intros. remember (eqX_dec x y) as b. destruct b.
  apply False_ind. apply H. assumption. eexists. reflexivity. Qed.

Definition contained (X : Type) (eqX_dec : forall x y : X, {x = y}+{x <> y}) (x : X) (xs : list X) :
  {In x xs} + {~In x xs}.
Proof.
  intros. induction xs. right. simpl. intro Hc. inversion Hc.
  simpl. destruct IHxs. left. right. assumption.
  specialize eqX_dec with (x := a) (y := x).
  destruct eqX_dec. left. left. assumption.
  right. apply not_or. split; assumption.
Defined.
Implicit Arguments contained [X].

Lemma contained_nil : forall X (x : X) eqX_dec, exists pf, contained eqX_dec x nil = right _ pf.
Proof. intros. simpl. eexists. reflexivity. Qed.

(* XXX could use in_right instead of existential
Program Lemma contained_nil : forall X (x : X) eqX_dec, contained eqX_dec x nil = in_right.
Proof. intros. simpl. reflexivity. Qed.
*)

Lemma not_in_contained_right : forall X (x : X) xs (eqX_dec :  forall x y : X, {x = y}+{x <> y}),
  ~ In x xs -> exists pf, contained eqX_dec x xs = right _ pf.
Proof. intros. induction xs. eexists. simpl. reflexivity.
  simpl in *. apply not_or_rev in H. destruct H as [H1 H2]. apply IHxs in H2.
  destruct H2 as [Hin H2]. rewrite H2.
  apply eqX_dec_false_rev with (eqX_dec := eqX_dec) in H1. destruct H1 as [H0 H1].
  rewrite H1. exists (not_or (a = x) (In x xs) (conj H0 Hin)). reflexivity.
Qed.

Lemma contained_monotonous : forall X (x : X) a1 a2 (eqX_dec :  forall x y : X, {x = y}+{x <> y}) pf2,
  contained eqX_dec x a2 = left _ pf2 ->
    exists pf12, contained eqX_dec x (a1 ++ a2) = left _ pf12.
Proof.
  intros. induction a1 as [| x1 a1']. simpl. exists pf2. assumption.
  simpl. destruct IHa1' as [pf12 IHa1']. rewrite IHa1'. exists (or_intror (x1 = x) pf12). reflexivity.
Qed.

Lemma not_in_app : forall X (z : X) xs ys,
  ~In z (xs ++ ys) <-> ~In z xs /\ ~In z ys.
Admitted.

Lemma map_cons : forall A B (x : A) xs (f : A -> B),
  map f (x :: xs) = f x :: map f xs.
Proof. trivial. Qed.

Lemma invert_app_last_eq : forall A (xs ys : list A) xn yn,
  xs ++ xn :: nil = ys ++ yn :: nil -> xn = yn.
Admitted.

(* Consecutive elemens of a list *)

Section In2.

  Variable A : Type.

  (* In2 x1 x2 xs means that x1 and x2 are consecutive elements in xs *)
  Fixpoint In2 (x1 x2 : A) (xs : list A) {struct xs} :=
    match xs with
    | y1 :: xs' => 
      (match xs' with
      | y2 :: xs'' => (x1 = y1 /\ x2 = y2)
      | _ => False
      end) \/ In2 x1 x2 xs'
    | _ => False
    end.

  Lemma in2_app : forall z1 z2 xs ys,
    In2 z1 z2 (xs ++ ys) ->
      In2 z1 z2 xs \/
      In2 z1 z2 ys \/
      (exists xs', exists ys',
        xs = xs' ++ (z1 :: nil) /\
        ys = z2 :: ys').
  Admitted.

  Lemma in2_app_middle : forall z1 z2 xs x ys,
    In2 z1 z2 (xs ++ x :: ys) ->
      In2 z1 z2 xs \/
      In2 z1 z2 ys \/
      (z2 = x /\ exists xs', xs = xs' ++ (z1 :: nil)) \/
      (z1 = x /\ exists ys', ys = z2 :: ys').
  Admitted.

End In2.
Implicit Arguments In2[A].

Lemma in2_map_cons : forall A (x : A) zs1 zs2 zss,
  In2 zs1 zs2 (map (cons x) zss) ->
    exists zs1', exists zs2',
      zs1 = x :: zs1' /\
      zs2 = x :: zs2' /\
      In2 zs1' zs2' zss.
Admitted.

Lemma in2_cons_cons_implies : forall Y (y1 y2 : Y) ys (P : Y -> Y -> Prop), 
  (forall y1' y2', In2 y1' y2' (y1 :: y2 :: ys) -> P y1' y2') ->
  P y1 y2 /\ (forall y1' y2', In2 y1' y2' (y2 :: ys) -> P y1' y2').
Admitted.

Lemma in2_cons_cons_implies_pairs : forall X Y (x1 x2 : X) (y1 y2 : Y) xys (P : X -> Y -> X -> Y -> Prop), 
  (forall x1' y1' x2' y2', In2 (x1',y1') (x2',y2') ((x1,y1) :: (x2,y2) :: xys) -> P x1' y1' x2' y2') ->
  P x1 y1 x2 y2 /\ (forall x1' y1' x2' y2', In2 (x1',y1') (x2',y2') ((x2,y2) :: xys) -> P x1' y1' x2' y2').
Admitted.


(* Splitting permutations as product of adjacent transpositions *)


Section AdjTrans.

  Variable A : Type.

  (* adjacent transpositions *)
  Inductive AdjTrans : list A -> list A -> Prop :=
    | adj_trans_skip: forall (x : A) (l l': list A), AdjTrans l l' -> AdjTrans (cons x l) (cons x l')
    | adj_trans_swap: forall (x y : A) (l : list A), AdjTrans (cons y (cons x l)) (cons x (cons y l)).

  Lemma adj_trans_perm : forall (xs ys : list A),
    AdjTrans xs ys -> Permutation xs ys.
  Proof. intros. induction H. apply perm_skip. assumption. apply perm_swap. Qed.

  Lemma prod_perm : forall (xs ys zs : list A),
    Permutation xs ys -> Permutation ys zs -> Permutation xs zs.
  Proof. apply Permutation_trans. Qed.

  (* Any permutation can be written as the product of adjacent transpositions *)
  Lemma split_perm_adj_trans : forall (xs ys : list A),
    Permutation xs ys ->
      xs = ys \/
      AdjTrans xs ys \/
        (exists zs0 : list A, exists zss' : list (list A),
          (* first step *)
          AdjTrans xs zs0 /\
          (* last step *)
          ((zss' = nil /\ AdjTrans zs0 ys) \/ 
            (exists zsn : list A, exists zss'',
              zss' = zss'' ++ (zsn :: nil) /\
              AdjTrans zsn ys)
          ) /\
          (* intermediate steps *)
          (forall zs1 zs2, In2 zs1 zs2 (zs0 :: zss') -> AdjTrans zs1 zs2)
        ).
  Proof.
    intros.
    induction H.
    Case "perm_nil". left. reflexivity.
    Case "perm_skip".
      destruct IHPermutation as [| IHPermutation].
      SCase "left". left. subst. reflexivity.
      SCase "right". right.
        destruct IHPermutation as [| IHPermutation].
        SSCase "left (AdjTrans)". left. apply adj_trans_skip. assumption.
        SSCase "right". right.
          destruct IHPermutation as [zs0 [zss' [IH1 [IH2 IH3]]]].
          exists (x :: zs0).
          exists (List.map (cons x) zss').
          split. apply adj_trans_skip. apply IH1.
          split. destruct IH2 as [[IH21 IH22] | [zsn [zss'' [IH21 IH22]]]].
            left. subst. simpl. split. trivial. apply adj_trans_skip. apply IH22.
            right. eexists. eexists. split.
              rewrite IH21. rewrite map_app. simpl. reflexivity.
              apply adj_trans_skip. apply IH22.
          intros zs1 zs2 H'.
            rewrite <- map_cons in H'.
            apply in2_map_cons in H'. destruct H' as [zs1' [zs2' [H1' [H2' H3']]]].
            subst. apply adj_trans_skip. apply IH3. apply H3'.
        SCase "perm_swap". right. left. apply adj_trans_swap.
        SCase "perm_trans".
          destruct IHPermutation1 as [| IHPermutation1];
          destruct IHPermutation2 as [| IHPermutation2].
          SSCase "left + left". subst. left. reflexivity.
          SSCase "left + right". subst. right. apply IHPermutation2.
          SSCase "right + left". subst. right. apply IHPermutation1.
          SSCase "right + right". right. right.
            destruct IHPermutation1 as [| IHPermutation1];
            destruct IHPermutation2 as [| IHPermutation2].
            SSSCase "left + left".
              exists l'. exists nil. split. assumption.
              split. left. split. reflexivity. assumption.
              intros. simpl in H3. destruct H3; inversion H3.
            SSSCase "left + right".
              destruct IHPermutation2 as [zs [zss' [IH21 [IH22 IH23]]]].
              exists l'. exists (zs :: zss').
              split. assumption.
              split. right. destruct IH22 as [ [IH' IH''] | ].
                subst. exists zs. exists nil. split. reflexivity. apply IH''.
                destruct H2 as [zsn [zss'' [H2' H2'']]]. subst.
                  exists zsn. exists (zs :: zss''). split. reflexivity. apply H2''.
              intros. simpl in H2. destruct H2 as [[H2' H2'']|].
                subst. assumption.
                apply IH23. apply H2.
            SSSCase "right + left (symmetric)". admit. (* symmetric *)
            SSSCase "right + right".
              destruct IHPermutation1 as [zs1 [zss'1 [IH11 [IH12 IH13]]]].
              destruct IHPermutation2 as [zs2 [zss'2 [IH21 [IH22 IH23]]]].
              exists zs1. exists (zss'1 ++ l' :: (zs2 :: zss'2)).
              split. assumption.
              split. right. destruct IH22 as [ [IH' IH''] | [zsn [zss' [IH' IH'']]]].
                subst. eexists. eexists. split.
                  assert(zss'1 ++ l' :: zs2 :: nil = (zss'1 ++ l' :: nil) ++ zs2 :: nil) as H1. admit. (* trivial *)
                  rewrite H1. reflexivity. assumption.
                subst. eexists zsn. eexists (zss'1 ++ l :: zs2 :: zss'). split. admit. (* trivial *) assumption.
              intros.
                assert ((zs1 :: zss'1 ++ l' :: zs2 :: zss'2) = ((zs1 :: zss'1) ++ l' :: (zs2 :: zss'2))) as H2. reflexivity.
                rewrite H2 in H1. clear H2.
                apply in2_app_middle in H1. destruct H1 as [|[|[|]]].
                SSSSCase "In2 zs0 zs3 (zs1 :: zss'1)". apply IH13. assumption.
                SSSSCase "In2 zs0 zs3 (zs2 :: zss'2)". apply IH23. assumption.
                SSSSCase "wow, one more case :)".
                  destruct H1. subst. destruct H2.
                  destruct IH12 as [ [IH' IH''] | [zsn [zss' [IH' IH'']]]].
                    subst. assert(zs1 :: nil = nil ++ zs1 :: nil) as H2. reflexivity.
                      rewrite H2 in H1. apply invert_app_last_eq in H1. subst. assumption.
                    subst.
                      assert(zs1 :: zss' ++ zsn :: nil = (zs1 :: zss') ++ zsn :: nil) as H2. reflexivity.
                      rewrite H2 in H1. apply invert_app_last_eq in H1. subst. assumption.
                SSSSCase "last one (symmetric)". admit. (* symmetric *)
  Qed.

End AdjTrans.

(* Stuff about last element of a list *)


Lemma last_cons_cons : forall X (x y d : X) xs,
  last (x :: y :: xs) d = last (y :: xs) d.
Proof. intros. reflexivity. Qed.

Lemma last_cons_change_default : forall X (xs : list X),
  xs <> nil ->
    forall d1 d2, List.last xs d1 = List.last xs d2.
Proof.
  intros. induction xs. apply False_ind. apply H. reflexivity.
  destruct xs. reflexivity. simpl in *. apply IHxs. intro Hc. inversion Hc.
Qed.

Lemma snd_last : forall X Y (xys : list (X * Y)) d,
  snd (last xys d) = last (snd (split xys)) (snd d).
Proof. intros. induction xys as [| [x y] xys']. reflexivity.
  destruct d as [dx dy]. destruct xys' as [| [x' y'] xys'']. reflexivity.
  simpl in *. rewrite IHxys'. destruct (split xys'').
  reflexivity.
Qed.

Lemma ex_intro_swap : forall (A : Type) (x : A) (P : A -> Prop), P x -> ex P.
Proof. intros. eapply ex_intro. eassumption. Qed.

