(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)

Require Import List.
Require Import Model.
Require Import Calculus.
Require Import LogicalSemantics.
Require Import Purity.

Definition T_Sing_T e T := 
  let (x, _) := fresh (fv_exp e) in
  if is_pure e then
    T_Refine x T (E_Eq (E_Var x) e)
  else T.
Definition T_Sing_av avoid e T := 
  let (x, _) := fresh (fv_exp e ++ avoid) in
  if is_pure e then
    T_Refine x T (E_Eq (E_Var x) e)
  else T.
Definition T_Sing_x x e T := if is_pure e then T_Refine x T (E_Eq (E_Var x) e) else T.
Definition T_Sing_Any e := T_Sing_T e T_Any.

Definition T_Ok e :=
  let (x, _) := fresh (fv_exp e) in
  if is_pure e then T_Refine x T_Any e else T_Any.
Definition T_Ok_av avoid e :=
  let (x, _) := fresh (fv_exp e ++ avoid) in
  if is_pure e then T_Refine x T_Any e else T_Any.
Definition T_Ok_x x e :=
  if is_pure e then T_Refine x T_Any e else T_Any.

Definition T_If (e : Exp) (T U : MType) :=
  let (x, _) := fresh (fv_exp e) in
  if is_pure e then T_Union (T_Refine x T e) (T_Refine x U (E_Not e)) else T_Union T U.


Lemma F_T_Sing_T : forall env e T t,
  is_pure e = true ->
    F (T_Sing_T e T) t env = (andb (syn_beq_res (R e env) (Return t)) (F T t env)).
Proof.
  intros env e T t H. unfold T_Sing_T. rewrite H.
  destruct (fresh (fv_exp e)) as [x H'].
  eapply R_not_in_fv_exp in H'. simpl. rewrite H'.
  remember (R e env) as r. destruct r; simpl; subst.
  Case "R e env = Error". destruct (F T t env); reflexivity.
  rewrite env_update_same.
  unfold O_EQ. rewrite syn_beq_val_v_logical_v_tt.
  rewrite syn_beq_val_sym.
  apply andb_comm.
Qed.

Lemma F_T_Sing_T_impure : forall env e T t,
  is_pure e = false -> F (T_Sing_T e T) t env = F T t env.
Proof. intros env e T t H. unfold T_Sing_T. rewrite H. destruct (fresh (fv_exp e)). reflexivity. Qed.

Lemma F_T_Ok : forall env e t,
  is_pure e = true -> F (T_Ok e) t env = (syn_beq_res (R e env) (Return v_tt)).
Proof.
  intros env e t H. unfold T_Ok. rewrite H. (*simpl. *)
  destruct (fresh (fv_exp e)) as [x H'].
  eapply R_not_in_fv_exp in H'. simpl. rewrite H'.
  reflexivity.
Qed.

Lemma F_T_Ok_impure : forall env e t,
  is_pure e = false -> F (T_Ok e) t env = true.
Proof. intros env e t H. unfold T_Ok. rewrite H. destruct (fresh (fv_exp e)). reflexivity. Qed.

Lemma F_T_If : forall env e T U t,
  is_pure e = true -> Proper (R e env) = true -> In_Logical (out_V (R e env)) = true ->
  (forall x, W (T_Refine x T e) t env = false) ->
  (forall x, W (T_Refine x U e) t env = false) ->
    F (T_If e T U) t env =
      orb (andb (syn_beq_res (R e env) (Return v_tt)) (F T t env))
          (andb (syn_beq_res (R e env) (Return v_ff)) (F U t env)).
Proof.
  intros env e T U t H1 H2 H3 H4 H5. unfold T_If. rewrite H1. 
  destruct (fresh (fv_exp e)) as [x H'].
  rewrite F_T_Union. simpl. eapply R_not_in_fv_exp in H'. rewrite H'. unfold LBind.
  rewrite H2. simpl. rewrite H3. rewrite O_Not_54. rewrite andb_comm.
  destruct (syn_beq_res (R e env) (Return v_tt) && F T t env); try reflexivity; simpl.
  rewrite andb_comm. rewrite syn_beq_val_syn_beq_res. rewrite Proper_out_V; try assumption.
  reflexivity. apply H4. apply W42. apply H5.
Qed.

Lemma F_T_If_impure : forall env e T U t,
  is_pure e = false -> F (T_If e T U) t env = F (T_Union T U) t env.
Proof. intros env e T U t H. unfold T_If. rewrite H. destruct (fresh (fv_exp e)). reflexivity. Qed.

